{{
***************************************************
* File jvmCompareJump: control transfer functions *
***************************************************
}}

{{
  Unconditional Branching. 
  One way to transfer control is with unconditional branching. The Virtual Machine handles this type of 
  branching with the bytecodes discussed in this section. 

  Conditional Branching. 
  Branching instructions check for a specific value. If the value is as expected, execution proceeds at an 
  offset from the address of the current instruction. Otherwise, execution proceeds to the next instruction. 
}}

{{
  if<cond>
  Operation: Branch if int comparison with zero succeeds
  Format: if<cond>,branchbyte1,branchbyte2   
  Operand Stack: ...,value --> ...,

  Description
  The value must be of type int. It is popped from the operand stack and compared against zero.
  All comparisons are signed. The results of the comparisons are as follows:

    eq succeeds if and only if value == 0 
    ne succeeds if and only if value <> 0 
    lt succeeds if and only if value  < 0 
    le succeeds if and only if value =< 0 
    gt succeeds if and only if value  > 0 
    ge succeeds if and only if value => 0 

  If the comparison succeeds, the unsigned branchbyte1 and branchbyte2 are used to construct a signed 16-bit
  offset, where the offset is calculated to be (branchbyte1 << 8) | branchbyte2. Execution then proceeds at
  that offset from the address of the opcode of this if<cond> instruction. The target address must be that
  of an opcode of an instruction within the method that contains this if<cond> instruction.
  Otherwise, execution proceeds at the address of the instruction following this if<cond> instruction.
}}

{{
  if_icmp<cond>
  Operation: Branch if int comparison succeeds
  Format: if_icmp<cond>,branchbyte1,branchbyte2   
  Operand Stack: ...,value1,value2 --> ...,

  Description
  Both value1 and value2 must be of type int. They are both popped from the operand stack and compared.
  All comparisons are signed. The results of the comparison are as follows:

    eq succeeds if and only if value1 == value2 
    ne succeeds if and only if value1 <> value2 
    lt succeeds if and only if value1  < value2 
    le succeeds if and only if value1 =< value2 
    gt succeeds if and only if value1  > value2 
    ge succeeds if and only if value1 => value2 

  If the comparison succeeds, the unsigned branchbyte1 and branchbyte2 are used to construct a signed 16-bit
  offset, where the offset is calculated to be (branchbyte1 << 8) | branchbyte2. Execution then proceeds at
  that offset from the address of the opcode of this if_icmp<cond> instruction. The target address must be
  that of an opcode of an instruction within the method that contains this if_icmp<cond> instruction.
  Otherwise, execution proceeds at the address of the instruction following this if_icmp<cond> instruction. 
}}

OBJ
  js: "jvmStack"
  jd: "jvmData"
  
  
PUB jb_goto | offset
{{ Meaning: Branch. }}
{{
  Description: Execution proceeds at the signed 16-bit offset from the address of this instruction.
  The offset is constructed from byte1 and byte2.
}}
{{
  JEM format: goto,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  jd.incJvmPC(offset)
  

{{PUB jb_goto_w}}
{{ Meaning: Branch always for wide index. }}
{{
  Description: Execution proceeds at the signed 32-bit offset from the address of this instruction.
  The offset is constructed from byte1, byte2, byte3 and byte4.
}}


PUB jb_ifeq | offset,w1
{{ Meaning: Branch if equal to 0. }}
{{
  JEM format: ifeq,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 == 0
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)
    

{{PUB jb_ifnull}}
{{ Meaning: Branch if null. }}


PUB jb_iflt | offset,w1
{{ Meaning: Branch if less than 0. }}
{{
  JEM format: iflt,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 < 0
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_ifle | offset,w1
{{ Meaning: Branch if less than or equal to 0. }}
{{
  JEM format: ifle,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 =< 0
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_ifne | offset,w1
{{ Meaning: Branch if not equal to 0. }}
{{
  JEM format: ifne,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 <> 0
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


{{PUB jb_ifnonnull}}
{{ Meaning: Branch if not null. }}


PUB jb_ifgt | offset,w1
{{ Meaning: Branch if greater than 0. }}
{{
  JEM format: ifgt,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 > 0
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_ifge | offset,w1
{{ Meaning: Branch if greater than or equal to 0. }}
{{
  JEM format: ifge,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 => 0
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_icmpeq | offset,w1,w2
{{ Meaning: Branch if integers v1 and v2 are equal. }}
{{
  JEM format: icmpeq,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 == w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_icmpne | offset,w1,w2
{{ Meaning: Branch if integers v1 and v2 are not equal. }}
{{
  JEM format: icmpne,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 <> w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_icmplt | offset,w1,w2
{{ Meaning: Branch if integer v1 is less than integer v2. }}
{{
  JEM format: icmplt,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 < w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_icmpgt | offset,w1,w2
{{ Meaning: Branch if integer v1 is greater than integer v2. }}
{{
  JEM format: icmpgt,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 > w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_icmple | offset,w1,w2
{{ Meaning: Branch if integer v1 is less than or equal to integer v2. }}
{{
  JEM format: icmple,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 =< w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_icmpge | offset,w1,w2
{{ Meaning: Branch if integer v1 is greater than or equal to integer v2. }}
{{
  JEM format: icmpge,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 => w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_acmpeq | offset,w1,w2
{{ Meaning: Branch if object references are equal. }}
{{
  JEM format: acmpeq,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 == w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


PUB jb_if_acmpne | offset,w1,w2
{{ Meaning: Branch if object references are not equal. }}
{{
  JEM format: acmpne,byte1,byte2
  byte1 and byte2 form 16bits offset to add to pc: offset = (byte1<<8)|(byte2) 
}}
  w2 := js.pop
  w2 := ~~w2 'sign extend b15
  w1 := js.pop
  w1 := ~~w1 'sign extend b15
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  if w1 <> w2
    jd.incJvmPC(offset)
  else
    jd.incJvmPC(3)


{{PUB jb_lcmp}}
{{ Meaning: Long integer compare. }}


{{PUB jb_fcmpl}}
{{ Meaning: Single-precision floating-point number compare; return -1 if v1 or v2 is not a number. }}


{{PUB jb_dcmpl}}
{{ Meaning: Double-precision floating-point number compare; return -1 if v1 or v2 is not a number. }}


{{PUB jb_fcmpg}}
{{ Meaning: Single-precision floating-point number compare; return +1 if v1 or v2 is not a number. }}


{{PUB jb_dcmpg}}
{{ Meaning: Double-precision floating-point number compare; return +1 if v1 or v2 is not a number. }}


PUB jb_tableswitch | default,low,high,index,offset
{{ Meaning: Access jump table by index and jump. }}
{{
  Description: Immediately after the tableswitch instruction, padding consisting of 0-3 zeros 
  is inserted so that the next byte begins at an address that is a multiple of four. A series of 
  offsets follows the padding. These offsets are signed 4-byte quantities and consist of a default 
  offset, a low offset, and a high offset, followed by additional <high-low+1> offsets. These 
  additional offsets are treated as a 0-based jump table.  

  The index the offsets point to must be an integer. If the index is less than the low offset or 
  greater than the high offset, the default offset is added to the address of the current 
  instruction. Otherwise, the low offset is subtracted from the index, and the element at the 
  position index - low offset in the jump table is extracted and added to the address of the 
  current instruction.
}}
{{
  JEM format: tableswitch,byte1,byte2,byte3,byte4,byte5,byte6,<number of offsets>
  default offset = (byte1<<8)|(byte2). Low value = (byte3<<8)|(byte4). High value - low value = (byte5<<8)|(byte6).
}}
  default := jd.readJavaIntBE(jd.getJvmPC+1)
  low := jd.readJavaIntBE(jd.getJvmPC+3)
  high := jd.readJavaIntBE(jd.getJvmPC+5) + low
  index := js.pop 'get index
  if index < low OR index > high
    jd.incJvmPC(default+1)
  else
    index -= low
    offset := jd.readJavaIntBE(jd.getJvmPC + 7 + (index<<1))
    jd.incJvmPC(offset+1)
  

PUB jb_lookupswitch | default,npairs,key,index,match,offset
{{ Meaning: Access jump table by key match and jump. }}
{{
  Description: Immediately after the lookupswitch instruction, padding consisting of 0-3 
  zeros is inserted so that the next byte begins at an address that is a multiple of four. A series 
  of pairs of offsets follows the padding. These offsets are signed 4-byte quantities. The first 
  item in the pair is the default offset, and the second item in the pair gives the number of pairs 
  that follow. The additional pairs consist of a match and an offset.  

  The integer key on the stack is then compared against each of the matches. If the key is equal 
  to one of the matches, the offset is added to the address of the current instruction. If the key  
  does not match any of the matches, the default offset is added to the address of the current instruction.
}}
{{
  JEM format: lookupswitch,byte1,byte2,byte3,byte4,<number of pairs: match,offset>
  default offset = (byte1<<8)|(byte2). Number of pairs = (byte3<<8)|(byte4).
}}
  default := jd.readJavaIntBE(jd.getJvmPC+1)
  npairs := jd.readJavaIntBE(jd.getJvmPC+3)
  key := js.pop 'get key
  index := 5
  repeat npairs
    match := jd.readJavaIntBE(jd.getJvmPC + index)
    offset := jd.readJavaIntBE(jd.getJvmPC + index + 2)
    if key == match
      jd.incJvmPC(offset+1)
      return
    index += 4
  jd.incJvmPC(default+1)
  
  