{{
**********************************************************
* File jvmVirtualPeripheral: native routines for VP code *
**********************************************************
}}

{{
  Virtual Peripherals are implemented as background code, meaning the code executes
  independantly from the interpreted java code.
}}


OBJ
  def:    "jvmDefines"
  vpCog1: "jvmVpCore"
  vpCog2: "jvmVpCore"

  
DAT
  vpTimerCount  long 0                       'holds 32bit free running timer, must be in front of vpRambank
  vpRambank     long 0[def#MAX_NUM_VPS*4+1]  'registers for Virtual Peripherals (16 registers per VP bank, starting at offset 2)

  
PUB start
{{ startup COG's for VP code }}
  vpTimerCount := 0 + 6          'first bank of registers for vpCog1 is at 6 bytes offset to vpTimerCount
  vpCog1.start(@vpTimerCount)
  vpTimerCount := 3<<4 + 6       'first bank of registers for vpCog2 is at 54 bytes offset to vpTimerCount
  vpCog2.start(@vpTimerCount)

  
PUB stop
{{ stop COG's for VP code }}
  vpCog1.stop
  vpCog2.stop

  
PUB writeVPregister(address,value)
{{ write value to vpRambank register }}
  byte[vpRambank + address - def#nmPeriph0Bank + 2] := value
  
  
PUB readVPregister(address):value
{{ read value from vpRambank register }}
  return byte[vpRambank + address - def#nmPeriph0Bank + 2] & $FF


PUB nm_core_Timer_init(bank)
  '/**
  ' * Initialize the 32bit counter. The 32bit nmTimer register should increment every 8.68 usec.
  ' * Called by Timer.start()
  ' *
  ' * @param bank The allocated area for the Timer registers.
  ' */
  vpTimerCount := 0
  writeVPregister(bank + def#nmPeriphISRVec, def#TIMER)
  

PUB nm_core_Timer_latch(bank) | latchedCount
  '/**
  ' * Latch the 32bit counter into the latch registers.
  ' * Called by Timer.tickLo()
  ' *
  ' * @param bank The allocated area for the Timer registers.
  ' */
  latchedCount := vpTimerCount                                  'get current count 
  writeVPregister(bank + def#nmLatchTimer1, latchedCount      ) 'write b0
  writeVPregister(bank + def#nmLatchTimer2, latchedCount >> 8 ) 'write b1
  writeVPregister(bank + def#nmLatchTimer3, latchedCount >> 16) 'write b2
  writeVPregister(bank + def#nmLatchTimer4, latchedCount >> 24) 'write b3
  

PUB nm_core_ADC_install(bank) | inmask,outmask
  '/**
  ' * Initialize an ADC as background task. Its parameters are in vpRambank.
  ' * Called by ADC.start()
  ' *
  ' * @param bank The allocated area for the ADC registers.
  ' */
  inmask := readVPregister(bank + def#nmInPin)
  writeVPregister(bank + def#nmAdcInMask + 0, 0)
  writeVPregister(bank + def#nmadcInMask + 1, 0)
  writeVPregister(bank + def#nmAdcInMask + 2, 0)
  writeVPregister(bank + def#nmAdcInMask + 3, 0)
  writeVPregister(bank + def#nmAdcInMask + readVPregister(bank + def#nmInPort + 2), inmask) 
  outmask := readVPregister(bank + def#nmOutPin + 2)  
  writeVPregister(bank + def#nmAdcOutMask + 0, 0)
  writeVPregister(bank + def#nmAdcOutMask + 1, 0)
  writeVPregister(bank + def#nmAdcOutMask + 2, 0)
  writeVPregister(bank + def#nmAdcOutMask + 3, 0)
  writeVPregister(bank + def#nmAdcOutMask + readVPregister(bank + def#nmOutPort + 2), outmask) 
  writeVPregister(bank + def#nmPeriphISRVec, def#ADC) 'start adc VP

  
PUB nm_core_DAC_install(bank) | outmask
  '/**
  ' * Initialize a DAC as background task. Its parameters are in vpRambank.
  ' * Called by DAC.start()
  ' *
  ' * @param bank The allocated area for the DAC registers.
  ' */
  outmask := readVPregister(bank + def#nmDacPin)  
  writeVPregister(bank + def#nmDacOutMask + 0, 0)
  writeVPregister(bank + def#nmDacOutMask + 1, 0)
  writeVPregister(bank + def#nmDacOutMask + 2, 0)
  writeVPregister(bank + def#nmDacOutMask + 3, 0)
  writeVPregister(bank + def#nmDacOutMask + readVPregister(bank + def#nmDacPort + 2), outmask) 
  writeVPregister(bank + def#nmDacAccu, 0)
  writeVPregister(bank + def#nmPeriphISRVec, def#DAC) 'start dac VP


PUB nm_core_PWM_install(bank) | outmask
  '/**
  ' * Initialize a PWM as background task. Its parameters are in vpRambank.
  ' * Called by PWM.start()
  ' *
  ' * @param bank The allocated area for the PWM registers.
  ' */
  outmask := readVPregister(bank + def#nmPWMPin)  
  writeVPregister(bank + def#nmPWMOutMask + 0, 0)
  writeVPregister(bank + def#nmPWMOutMask + 1, 0)
  writeVPregister(bank + def#nmPWMOutMask + 2, 0)
  writeVPregister(bank + def#nmPWMOutMask + 3, 0)
  writeVPregister(bank + def#nmPWMOutMask + readVPregister(bank + def#nmPWMPort), outmask) 
  writeVPregister(bank + def#nmPeriphISRVec, def#PWM) 'start pwm VP


PUB nm_core_PWM_updateInternal(bank,highTime,lowTime,highZero,lowZero)
  '/**
  ' * Update the PWM duty cycle. Its parameters are in vpRambank. 
  ' * Called by PWM.update()
  ' *
  ' * @param bank The allocated area for the PWM registers.
  ' * @param highTime the number of 8.68us the waveform should be high. (1-255)
  ' * @param lowTime the number of 8.68us the waveform should be low. (1-255)
  ' * @param highZero Boolean value (always true) that tells wether highTime != 0.
  ' * @param lowZero Boolean value (always true) that tells wether lowTime != 0.
  ' */
  writeVPregister(bank + def#nmPWMlowTime, lowTime)
  writeVPregister(bank + def#nmPWMhighTime, highTime)
  writeVPregister(bank + def#nmPWMlowZero, lowZero)
  writeVPregister(bank + def#nmPWMhighZero, highZero)
       

PUB nm_core_Uart_rxInit(bank)
  '/**
  ' * Initialize a Uart receiver as background task. Its parameters are in vpRambank.
  ' * Called by Uart.start()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  writeVPregister(bank + def#nmPeriphISRVec, def#UARTRX)


PUB nm_core_Uart_rxRead(bank) | hsPortPin,head,tail,len
  '/**
  ' * Inform Uart receiver a byte has been grabbed from the buffer, so the receiver can do the handshaking.
  ' * Called by Uart.readByte()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  hsPortPin := (readVPregister(bank + def#nmUartHSPort) << 8) | readVPregister(bank + def#nmUartHSPin)
  if hsPortPin <> 0
    head := readVPregister(bank + def#nmUartHead)
    tail := readVPregister(bank + def#nmUartTail)
    len := head - tail
    if len < 0
      len += 256
    if len =< 16
      'enable hsPortpin (tell remote sender to resume transmitting)
    if len => 240
      'disable hsPortPin (tell remote sender to stop transmitting)

      
PUB nm_core_Uart_txInit(bank)
  '/**
  ' * Initialize a Uart transmitter as background task. Its parameters are in vpRambank.
  ' * Called by Uart.start()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  writeVPregister(bank + def#nmPeriphISRVec, def#UARTTX)

  